@@ -16,10 +16,13 @@
 #include "disjuncts.h"
 #include "error.h"
 #include "preparation.h"
+#include "read-regex.h"
+#include "regex-morph.h"
 #ifdef USE_SAT_SOLVER
 #include "sat-solver/sat-encoder.h"
 #endif
 #include "corpus/corpus.h"
+#include "spellcheck-hun.h"
 
 /***************************************************************
 *
@@ -76,6 +79,7 @@ Parse_Options parse_options_create(void)
 	po->max_null_count = 0;
 	po->null_block = 1;
 	po->islands_ok = FALSE;
+	po->use_spell_guess = TRUE;
 	po->cost_model.compare_fn = &VDAL_compare_parse;
 	po->cost_model.type = VDAL;
 	po->short_length = 6;
@@ -191,6 +195,14 @@ int parse_options_get_islands_ok(Parse_Options opts) {
 	return opts->islands_ok;
 }
 
+void parse_options_set_spell_guess(Parse_Options opts, int dummy) {
+	opts->use_spell_guess = dummy;
+}
+
+int parse_options_get_spell_guess(Parse_Options opts) {
+	return opts->use_spell_guess;
+}
+
 void parse_options_set_short_length(Parse_Options opts, int short_length) {
 	opts->short_length = short_length;
 }
@@ -503,10 +515,12 @@ static void affix_list_delete(Dictionary dict)
  * because an affix_file is opened by a recursive call to this
  * function.
  */
-Dictionary
-dictionary_create(const char * dict_name, const char * pp_name,
-                  const char * cons_name, const char * affix_name)
+static Dictionary
+dictionary_six(const char * lang, const char * dict_name,
+                const char * pp_name, const char * cons_name,
+                const char * affix_name, const char * regex_name)
 {
+	const char * t;
 	Dictionary dict;
 	Dict_node *dict_node;
 
@@ -517,7 +531,13 @@ dictionary_create(const char * dict_name, const char * pp_name,
 	memset(dict, 0, sizeof(struct Dictionary_s));
 
 	dict->string_set = string_set_create();
+	
+	dict->lang = lang;
+	t = strrchr (lang, '/');
+	if (t) dict->lang = string_set_add(t+1, dict->string_set);
 	dict->name = string_set_add(dict_name, dict->string_set);
+
+	dict->max_cost = 1000;
 	dict->num_entries = 0;
 	dict->is_special = FALSE;
 	dict->already_got_it = '\0';
@@ -528,13 +548,18 @@ dictionary_create(const char * dict_name, const char * pp_name,
 	dict->affix_table = NULL;
 	dict->recursive_error = FALSE;
 
+	/* To disable spell-checking, just set the cheker to NULL */
+	dict->spell_checker = spellcheck_create(dict->lang);
+
 	dict->fp = dictopen(dict->name, "r");
-	if (dict->fp == NULL) {
+	if (dict->fp == NULL)
+	{
 		prt_error("Error: Could not open dictionary %s\n", dict_name);
 		goto failure;
 	}
 
-	if (!read_dictionary(dict)) {
+	if (!read_dictionary(dict))
+	{
 		fclose(dict->fp);
 		goto failure;
 	}
@@ -543,7 +568,7 @@ dictionary_create(const char * dict_name, const char * pp_name,
 	dict->affix_table = NULL;
 	if (affix_name != NULL)
 	{
-		dict->affix_table = dictionary_create(affix_name, NULL, NULL, NULL);
+		dict->affix_table = dictionary_six(lang, affix_name, NULL, NULL, NULL, NULL);
 		if (dict->affix_table == NULL)
 		{
 			goto failure;
@@ -551,6 +576,24 @@ dictionary_create(const char * dict_name, const char * pp_name,
 		affix_list_create(dict->affix_table);
 	}
 
+	dict->regex_root = NULL;
+	if (regex_name != NULL)
+	{
+		int rc;
+		rc = read_regex_file(dict, regex_name);
+		if (rc) goto failure;
+		rc = compile_regexs(dict);
+		if (rc) goto failure;
+	}
+
+#if USE_CORPUS
+	dict->corpus = NULL;
+	if (affix_name != NULL) /* Don't do this for the second time */
+	{
+		dict->corpus = lg_corpus_new();
+	}
+#endif
+
 	dict->left_wall_defined  = boolean_dictionary_lookup(dict, LEFT_WALL_WORD);
 	dict->right_wall_defined = boolean_dictionary_lookup(dict, RIGHT_WALL_WORD);
 	dict->postprocessor	  = post_process_open(pp_name);
@@ -558,15 +601,19 @@ dictionary_create(const char * dict_name, const char * pp_name,
 
 	dict->unknown_word_defined = boolean_dictionary_lookup(dict, UNKNOWN_WORD);
 	dict->use_unknown_word = TRUE;
+
+#if DONT_USE_REGEX_GUESSING
 	dict->capitalized_word_defined = boolean_dictionary_lookup(dict, PROPER_WORD);
 	dict->pl_capitalized_word_defined = boolean_dictionary_lookup(dict, PL_PROPER_WORD);
+
 	dict->hyphenated_word_defined = boolean_dictionary_lookup(dict, HYPHENATED_WORD);
 	dict->number_word_defined = boolean_dictionary_lookup(dict, NUMBER_WORD);
+
 	dict->ing_word_defined = boolean_dictionary_lookup(dict, ING_WORD);
 	dict->s_word_defined = boolean_dictionary_lookup(dict, S_WORD);
 	dict->ed_word_defined = boolean_dictionary_lookup(dict, ED_WORD);
 	dict->ly_word_defined = boolean_dictionary_lookup(dict, LY_WORD);
-	dict->max_cost = 1000;
+#endif /* DONT_USE_REGEX_GUESSING */
 
 	if ((dict_node = dictionary_lookup_list(dict, ANDABLE_CONNECTORS_WORD)) != NULL) {
 		dict->andable_connector_set = connector_set_create(dict_node->exp);
@@ -590,28 +637,43 @@ failure:
 	return NULL;
 }
 
+Dictionary
+dictionary_create(const char * dict_name, const char * pp_name,
+                  const char * cons_name, const char * affix_name)
+{
+	return dictionary_six("en", dict_name, pp_name, cons_name, affix_name, NULL);
+}
+
 Dictionary dictionary_create_lang(const char * lang)
 {
 	Dictionary dictionary;
 
-	if(lang && *lang) {
+	if(lang && *lang)
+	{
 		char * dict_name;
 		char * pp_name;
 		char * cons_name;
 		char * affix_name;
+		char * regex_name;
 	
 		dict_name = join_path(lang, "4.0.dict");
 		pp_name = join_path(lang, "4.0.knowledge");
 		cons_name = join_path(lang, "4.0.constituent-knowledge");
 		affix_name = join_path(lang, "4.0.affix");
+		regex_name = join_path(lang, "4.0.regex");
 	
-		dictionary = dictionary_create(dict_name, pp_name, cons_name, affix_name);
+		dictionary = dictionary_six(lang, dict_name, pp_name, cons_name,
+		                             affix_name, regex_name);
 	
+		free(regex_name);
 		free(affix_name);
 		free(cons_name);
 		free(pp_name);
 		free(dict_name);
-	} else {
+	}
+	else
+	{
+		prt_error("Error: No language specified!\n");
 		dictionary = NULL;
 	}
 
@@ -641,10 +703,15 @@ int dictionary_delete(Dictionary dict)
 		prt_error("Info: Freeing dictionary %s\n", dict->name);
 	}
 
+#if USE_CORPUS
+	lg_corpus_delete(dict->corpus);
+#endif
+
 	if (dict->affix_table != NULL) {
 		affix_list_delete(dict->affix_table);
 		dictionary_delete(dict->affix_table);
 	}
+	spellcheck_destroy(dict->spell_checker);
 
 	connector_set_delete(dict->andable_connector_set);
 	connector_set_delete(dict->unlimited_connector_set);
@@ -652,6 +719,7 @@ int dictionary_delete(Dictionary dict)
 	post_process_close(dict->postprocessor);
 	post_process_close(dict->constituent_pp);
 	string_set_delete(dict->string_set);
+	free_regexs(dict);
 	free_dictionary(dict);
 	xfree(dict, sizeof(struct Dictionary_s));
 
@@ -733,9 +801,6 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 	int overflowed, only_canonical_allowed;
 	Linkage_info *link_info;
 	int canonical;
-#ifdef USE_CORPUS
-	Corpus * corp;
-#endif
 
 	free_post_processing(sent);
 
@@ -840,10 +905,6 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 		}
 	}
 
-#ifdef USE_CORPUS
-	corp = lg_corpus_new();
-#endif
-
 	/* second pass: actually perform post-processing */
 	N_linkages_post_processed = 0;
 	for (in=0; (in < N_linkages_alloced) &&
@@ -870,16 +931,10 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 			N_valid_linkages++;
 		}
 		lifo->index = indices[in];
-#ifdef USE_CORPUS
-		lg_corpus_score(corp, sent, lifo);
-#endif
+		lg_corpus_score(sent, lifo);
 		N_linkages_post_processed++;
 	}
 
-#ifdef USE_CORPUS
-	lg_corpus_delete(corp);
-#endif
-
 	print_time(opts, "Postprocessed all linkages");
 	qsort((void *)link_info, N_linkages_post_processed, sizeof(Linkage_info),
 		  (int (*)(const void *, const void *)) opts->cost_model.compare_fn);
@@ -981,11 +1036,6 @@ Sentence sentence_create(const char *input_string, Dictionary dict)
 		}
 	}
 
-	if (!build_sentence_expressions(sent)) {
-		sentence_delete(sent);
-		return NULL;
-	}
-
 	return sent;
 }
 
@@ -1078,6 +1128,14 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 
 	verbosity = opts->verbosity;
 
+	/* Tokenize, look up each word in the dictionary, collect up all
+	 * plausible disjunct expressions for each word. */
+	if (!build_sentence_expressions(sent, opts)) {
+		sent->num_valid_linkages = 0;
+		return 0;
+	}
+
+	/* Initialize/free any leftover garbage */
 	free_sentence_disjuncts(sent);
 	resources_reset_space(opts->resources);
 
@@ -1089,11 +1147,14 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 	init_analyze(sent);
 	init_count(sent);
 
+	/* Expressions were previously set up during the tokenize stage. */
 	expression_prune(sent);
 	print_time(opts, "Finished expression pruning");
 #ifdef USE_SAT_SOLVER
 	sat_parse(sent, opts);
 #else
+
+	/* Build lists of disjuncts */
 	prepare_to_parse(sent, opts);
 
 	init_fast_matcher(sent);
@@ -1109,7 +1170,7 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 	{
 		if (resources_exhausted(opts->resources)) break;
 		sent->null_count = nl;
-		total = parse(sent, sent->null_count, opts);
+		total = do_parse(sent, sent->null_count, opts);
 
 		/* Give up if the parse count is overflowing */
 		if (PARSE_NUM_OVERFLOW < total) break;
