@@ -1,6 +1,7 @@
 /*************************************************************************/
 /* Copyright (c) 2004                                                    */
 /* Daniel Sleator, David Temperley, and John Lafferty                    */
+/* Copyright (c) 2009 Linas Vepstas                                      */
 /* All rights reserved                                                   */
 /*                                                                       */
 /* Use of the link grammar parsing system is subject to the terms of the */
@@ -15,8 +16,11 @@
 #include <langinfo.h>
 #endif
 #include <limits.h>
+
 #include <link-grammar/api.h>
 #include "error.h"
+#include "regex-morph.h"
+#include "spellcheck-hun.h"
 #include "utilities.h"
 
 #define MAX_STRIP 10
@@ -27,7 +31,8 @@
 /*static char * strip_right[] = {")", "%", ",", ".", ":", ";", "?", "!", "''", "'", "'s", NULL};*/
 
 /**
- * This is rather esoteric and not terribly important.
+ * Is the word entirely composed of single-letter abreviations
+ * (followed by a period)?
  * It returns TRUE if the word matches the pattern /[A-Z]\.]+/
  */
 static int is_initials_word(const char * word)
@@ -68,6 +73,7 @@ static int is_number(const char * s)
 	return TRUE;
 }
 
+#if DONT_USE_REGEX_GUESSING
 /**
  * Returns TRUE iff it's an appropriately formed hyphenated word.
  * This means all letters, numbers, or hyphens, not beginning and
@@ -95,9 +101,10 @@ static int ishyphenated(const char * s)
 }
 
 /** 
- * The following four routines implement a cheap-hack morphology for
- * English.
- * XXX this is wrong for non-english
+ * The following four routines implement a cheap-hack morphology 
+ * guessing of unkown words for English.
+ * XXX Now obsolete, replaced by regex-based guessing.
+ * Remove dead code at liesure.
  */
 static int is_ing_word(const char * s)
 {
@@ -135,6 +142,7 @@ static int is_ly_word(const char * s)
 	if (strncmp("ly", s-2, 2)==0) return TRUE;
 	return FALSE;
 }
+#endif /* DONT_USE_REGEX_GUESSING */
 
 /** 
  * The string s is the next word of the sentence. 
@@ -236,12 +244,14 @@ static int downcase_is_in_dict(Dictionary dict, char * word)
 	char save[MB_LEN_MAX];
 	wchar_t c;
 	int nbl, nbh;
+	mbstate_t mbss;
 
 	if (!is_utf8_upper(word)) return FALSE;
 
 	nbh = mbtowc (&c, word, MB_CUR_MAX);
 	c = towlower(c);
-	nbl = wctomb_check(low, c);
+	nbl = wcrtomb(NULL, L'\0', &mbss);
+	nbl = wctomb_check(low, c, &mbss);
 	if (nbh != nbl)
 	{
 		prt_error("Warning: can't downcase multi-byte string: %s\n", word);
@@ -265,17 +275,20 @@ static int downcase_is_in_dict(Dictionary dict, char * word)
  * parts.  The process is described above.  Returns TRUE if OK, FALSE if
  * too many punctuation marks 
  */
-static int separate_word(Sentence sent, char *w, char *wend, int is_first_word, int quote_found)
+static int separate_word(Sentence sent, 
+                         const char *w, const char *wend,
+                         int is_first_word, int quote_found)
 {
 	int i, j, len;
 	int r_strippable=0, l_strippable=0, u_strippable=0;
 	int s_strippable=0, p_strippable=0;
 	int  n_r_stripped, s_stripped;
 	int word_is_in_dict, s_ok;
+	int issued = FALSE;
 
 	int found_number = 0;
 	int n_r_stripped_save;
-	char * wend_save;
+	const char * wend_save;
 
 	const char ** strip_left = NULL;
 	const char ** strip_right = NULL;
@@ -406,16 +419,16 @@ static int separate_word(Sentence sent, char *w, char *wend, int is_first_word,
 	s_stripped = -1;
 	strncpy(word, w, MIN(wend-w, MAX_WORD));
 	word[MIN(wend-w, MAX_WORD)] = '\0';
-	word_is_in_dict = 0;
+	word_is_in_dict = FALSE;
 
 	if (boolean_dictionary_lookup(sent->dict, word))
-		word_is_in_dict = 1;
+		word_is_in_dict = TRUE;
 	else if (is_initials_word(word))
-		word_is_in_dict = 1;
+		word_is_in_dict = TRUE;
 	else if (is_first_word && downcase_is_in_dict (sent->dict,word))
-		word_is_in_dict = 1;
+		word_is_in_dict = TRUE;
 
-	if (0 == word_is_in_dict)
+	if (FALSE == word_is_in_dict)
 	{
 		j=0;
 		for (i=0; i <= s_strippable; i++)
@@ -449,6 +462,7 @@ static int separate_word(Sentence sent, char *w, char *wend, int is_first_word,
 					wend -= len;
 					strncpy(word, w, MIN(wend-w, MAX_WORD));
 					word[MIN(wend-w, MAX_WORD)] = '\0';
+					word_is_in_dict = TRUE;
 					break;
 				}
 
@@ -475,6 +489,7 @@ static int separate_word(Sentence sent, char *w, char *wend, int is_first_word,
 								sz = MIN(wend-w, MAX_WORD);
 								strncpy(word, w, sz);
 								word[sz] = '\0';
+								word_is_in_dict = TRUE;
 								break;
 							}
 						}
@@ -487,7 +502,7 @@ static int separate_word(Sentence sent, char *w, char *wend, int is_first_word,
 	
 	/* word is now what remains after all the stripping has been done */
 
-	if (n_r_stripped == MAX_STRIP)
+	if (n_r_stripped >= MAX_STRIP)
 	{
 		prt_error("Error separating sentence.\n"
 		          "\"%s\" is followed by too many punctuation marks.\n", word);
@@ -496,9 +511,48 @@ static int separate_word(Sentence sent, char *w, char *wend, int is_first_word,
 
 	if (quote_found == TRUE) sent->post_quote[sent->length]=1;
 
-	if (!issue_sentence_word(sent, word)) return FALSE;
+	/* If the word is still not being found, then it might be 
+	 * a run-on of two words. Ask the spell-checker to split
+	 * the word in two, if possible.
+	 */
+	issued = FALSE;
+	if ((FALSE == word_is_in_dict) && sent->dict->spell_checker)
+	{
+		char **alternates = NULL;
+		char *sp = NULL;
+		char *wp;
+		int j, n;
+		n = spellcheck_suggest(sent->dict->spell_checker, &alternates, word);
+		for (j=0; j<n; j++)
+		{
+			/* Uhh, XXX this is not utf8 safe! */
+			sp = strchr(alternates[j], ' ');
+			if (sp) break;
+		}
 
-	if(s_stripped != -1)
+		if (sp) issued = TRUE;
+
+		wp = alternates[j];
+		while (sp)
+		{
+			*sp = 0x0;
+			issue_sentence_word(sent, wp);
+			wp = sp+1;
+			sp = strchr(wp, ' ');
+			if (NULL == sp)
+			{
+				issue_sentence_word(sent, wp);
+			}
+		}
+		if (alternates) free(alternates);
+	}
+
+	if (FALSE == issued)
+	{
+		if (!issue_sentence_word(sent, word)) return FALSE;
+	}
+
+	if (s_stripped != -1)
 	{
 	  if (!issue_sentence_word(sent, suffix[s_stripped])) return FALSE;
 	}
@@ -581,16 +635,21 @@ failure:
 	return FALSE;
 }
 
+#if DONT_USE_REGEX_GUESSING
 static int special_string(Sentence sent, int i, const char * s)
 {
 	X_node * e;
-	if (boolean_dictionary_lookup(sent->dict, s)) {
+	if (boolean_dictionary_lookup(sent->dict, s))
+	{
 		sent->word[i].x = build_word_expressions(sent, s);
-		for (e = sent->word[i].x; e != NULL; e = e->next) {
+		for (e = sent->word[i].x; e != NULL; e = e->next)
+		{
 			e->string = sent->word[i].string;
 		}
 		return TRUE;
-	} else {
+	}
+	else
+	{
 		err_ctxt ec;
 		ec.sent = sent;
 		err_msg(&ec, Error, "Error: Could not build sentence expressions.\n"
@@ -599,7 +658,42 @@ static int special_string(Sentence sent, int i, const char * s)
 		return FALSE;
 	}
 }
+#endif /* DONT_USE_REGEX_GUESSING */
 
+/**
+ * Build the word expressions, and add a tag to the word to indicate
+ * that it was guess by means of regular-expression matching.
+ * Also, add a subscript to the resulting word to indicate the
+ * rule origin.
+ */
+static void tag_regex_string(Sentence sent, int i, const char * type)
+{
+	char str[MAX_WORD+1];
+	char * t;
+	X_node * e;
+	sent->word[i].x = build_word_expressions(sent, type);
+	for (e = sent->word[i].x; e != NULL; e = e->next)
+	{
+		t = strchr(e->string, '.');
+		e->string = sent->word[i].string;
+		if (NULL != t)
+		{
+			snprintf(str, MAX_WORD, "%.50s[!].%.5s", e->string, t+1);
+		}
+		else
+		{
+			snprintf(str, MAX_WORD, "%.50s", e->string);
+		}
+		e->string = string_set_add(str, sent->string_set);
+	}
+}
+
+#if DONT_USE_REGEX_GUESSING
+/**
+ * Performes morphology-based word guessing.
+ * Obsolete with the new regex infratructure ... remove this code at
+ * liesure.
+ */
 static int guessed_string(Sentence sent, int i, const char * s, const char * type)
 {
 	X_node * e;
@@ -639,8 +733,9 @@ static int guessed_string(Sentence sent, int i, const char * s, const char * typ
 			{
 				sprintf(str, "%.50s[!].e", s);
 			}
-			else sprintf(str, "%.50s[!]", s);
+			else
 
+			sprintf(str, "%.50s[!]", s);
 			e->string = string_set_add(str, sent->string_set);
 		}
 		return TRUE;
@@ -656,6 +751,7 @@ static int guessed_string(Sentence sent, int i, const char * s, const char * typ
 		return FALSE;
 	}
 }
+#endif /* DONT_USE_REGEX_GUESSING */
 
 /**
  * Puts into word[i].x the expression for the unknown word 
@@ -679,38 +775,108 @@ static void handle_unknown_word(Sentence sent, int i, char * s)
 		t = strchr(d->string, '.');
 		if (t != NULL)
 		{
-			sprintf(str, "%.50s[?].%.5s", s, t+1);
+			snprintf(str, MAX_WORD, "%.50s[?].%.5s", s, t+1);
 		}
 		else
 		{
-			sprintf(str, "%.50s[?]", s);
+			snprintf(str, MAX_WORD, "%.50s[?]", s);
 		}
 		d->string = string_set_add(str, sent->string_set);
 	}
 }
 
+/**
+ * If a word appears to be mis-spelled, then add alternate
+ * spellings. Maybe one of those will do ... 
+ */
+static void guess_misspelled_word(Sentence sent, int i, char * s)
+{
+	int spelling_ok;
+	char str[MAX_WORD+1];
+	Dictionary dict = sent->dict;
+	X_node *d, *head = NULL;
+	int j, n;
+	char **alternates = NULL;
+
+	/* Spell-guessing is disabled if no spell-checker is speficified */
+	if (NULL == dict->spell_checker)
+	{
+		handle_unknown_word(sent, i, s);
+		return;
+	}
+
+	/* If the spell-checker knows about this word, and we don't ... 
+	 * Dang. We should fix it  someday. Accept it as such. */
+	spelling_ok = spellcheck_test(dict->spell_checker, s);
+	if (spelling_ok)
+	{
+		handle_unknown_word(sent, i, s);
+		return;
+	}
+
+	/* Else, ask the spell-checker for alternate spellings
+	 * and see if these are in the dict. */
+	n = spellcheck_suggest(dict->spell_checker, &alternates, s);
+	for (j=0; j<n; j++)
+	{
+		if (boolean_dictionary_lookup(sent->dict, alternates[j]))
+		{
+			X_node *x = build_word_expressions(sent, alternates[j]);
+			head = catenate_X_nodes(x, head);
+		}
+	}
+	sent->word[i].x = head;
+	if (alternates) free(alternates);
+
+	/* Add a [~] to the output to signify that its the result of
+	 * guessing. */
+	for (d = sent->word[i].x; d != NULL; d = d->next)
+	{
+		const char * t = strchr(d->string, '.');
+		if (t != NULL)
+		{
+			size_t off = t - d->string;
+			strncpy(str, d->string, off);
+			str[off] = 0;
+			strcat(str, "[~]");
+			strcat(str, t);
+		}
+		else
+		{
+			snprintf(str, MAX_WORD, "%.50s[~]", s);
+		}
+		d->string = string_set_add(str, sent->string_set);
+	}
+
+	/* If nothing found at all... */
+	if (NULL == head)
+	{
+		handle_unknown_word(sent, i, s);
+	}
+}
+
 /**
  * Corrects case of first word, fills in other proper nouns, and
  * builds the expression lists for the resulting words.
  *
  * Algorithm:
  * Apply the following step to all words w:
- * if w is in the dictionary, use it.
- * else if w is upper case use PROPER_WORD disjuncts for w.
- * else if it's hyphenated, use HYPHENATED_WORD
- * else if it's a number, use NUMBER_WORD.
+ * If w is in the dictionary, use it.
+ * Else if w is identified by regex matching, use the
+ * appropriately matched disjunct collection.
  *
  * Now, we correct the first word, w.
- * if w is upper case, let w' be the lower case version of w.
- * if both w and w' are in the dict, concatenate these disjncts.
- * else if w' is in dict, use disjuncts of w'
- * else leave the disjuncts alone
+ * If w is upper case, let w' be the lower case version of w.
+ * If both w and w' are in the dict, concatenate these disjncts.
+ * Else if w' is in dict, use disjuncts of w',
+ * else leave the disjuncts alone.
  */
-int build_sentence_expressions(Sentence sent)
+int build_sentence_expressions(Sentence sent, Parse_Options opts)
 {
 	int i, first_word;  /* the index of the first word after the wall */
 	char *s, temp_word[MAX_WORD+1];
 	const char * u;
+	const char * regex_name;
 	X_node * e;
 	Dictionary dict = sent->dict;
 
@@ -720,8 +886,8 @@ int build_sentence_expressions(Sentence sent)
 		first_word = 0;
 	}
 
-	/* the following loop treats all words the same
-	   (nothing special for 1st word) */
+	/* The following loop treats all words the same
+	 * (nothing special for 1st word) */
 	for (i=0; i<sent->length; i++)
 	{
 		s = sent->word[i].string;
@@ -729,6 +895,7 @@ int build_sentence_expressions(Sentence sent)
 		{
 			sent->word[i].x = build_word_expressions(sent, s);
 		}
+#if DONT_USE_REGEX_GUESSING
 		else if (is_utf8_upper(s) && is_s_word(s) && dict->pl_capitalized_word_defined) 
 		{
 			if (!special_string(sent, i, PL_PROPER_WORD)) return FALSE;
@@ -737,6 +904,13 @@ int build_sentence_expressions(Sentence sent)
 		{
 			if (!special_string(sent, i, PROPER_WORD)) return FALSE;
 		}
+#endif /* DONT_USE_REGEX_GUESSING */
+		else if ((NULL != (regex_name = match_regex(sent->dict, s))) &&
+		         boolean_dictionary_lookup(sent->dict, regex_name))
+		{
+			tag_regex_string(sent, i, regex_name);
+		}
+#if DONT_USE_REGEX_GUESSING
 		else if (is_number(s) && dict->number_word_defined)
 		{
 			/* we know it's a plural number, or 1 */
@@ -750,9 +924,9 @@ int build_sentence_expressions(Sentence sent)
 		} 
 		/* XXX
 		 * The following does some morphology-guessing for words that
-		 * that are not in the dictionary. This should be replaced by
-		 * a generic morphology-guesser for langauges that aren't english.
-		 * XXX
+		 * that are not in the dictionary. This has been replaced by a
+		 * regex-based morphlogy guesser. Remove this obsolete code at
+		 * liesure.
 		 */
 		else if (is_ing_word(s) && dict->ing_word_defined) 
 		{
@@ -770,9 +944,18 @@ int build_sentence_expressions(Sentence sent)
 		{
 			if (!guessed_string(sent, i, s, LY_WORD)) return FALSE;
 		}
+#endif /* DONT_USE_REGEX_GUESSING */
+
 		else if (dict->unknown_word_defined && dict->use_unknown_word)
 		{
-			handle_unknown_word(sent, i, s);
+			if (opts->use_spell_guess)
+			{
+				guess_misspelled_word(sent, i, s);
+			}
+			else
+			{
+				handle_unknown_word(sent, i, s);
+			}
 		}
 		else 
 		{
@@ -848,7 +1031,11 @@ int sentence_in_dictionary(Sentence sent)
 	for (w=0; w<sent->length; w++)
 	{
 		s = sent->word[w].string;
-		if (!boolean_dictionary_lookup(dict, s) &&
+		/* xx we should also check if the regex rules know this word */
+		/* However, this code is only used if the 'unknown word' is 
+		 * turned off ... we punt for now, since its always on. */
+		if (!boolean_dictionary_lookup(dict, s))
+#if DONT_USE_REGEX_GUESSING
 		    !(is_utf8_upper(s)   && dict->capitalized_word_defined) &&
 		    !(is_utf8_upper(s) && is_s_word(s) && dict->pl_capitalized_word_defined) &&
 		    !(ishyphenated(s) && dict->hyphenated_word_defined)  &&
@@ -857,6 +1044,7 @@ int sentence_in_dictionary(Sentence sent)
 		    !(is_s_word(s)	&& dict->s_word_defined)  &&
 		    !(is_ed_word(s)   && dict->ed_word_defined)  &&
 		    !(is_ly_word(s)   && dict->ly_word_defined))
+#endif /* DONT_USE_REGEX_GUESSING */
 		{
 			if (ok_so_far)
 			{
