@@ -162,10 +162,12 @@ void downcase_utf8_str(char *to, const char * from, size_t usize)
 	wchar_t c;
 	int i, nbl, nbh;
 	char low[MB_LEN_MAX];
+	mbstate_t mbss;
 
 	nbh = mbtowc (&c, from, MB_CUR_MAX);
 	c = towlower(c);
-	nbl = wctomb_check(low, c);
+	nbl = wcrtomb(NULL, L'\0', &mbss);
+	nbl = wctomb_check(low, c, &mbss);
 
 	/* Check for error on an in-place copy */
 	if ((nbh < nbl) && (to == from))
@@ -193,10 +195,12 @@ void upcase_utf8_str(char *to, const char * from, size_t usize)
 	wchar_t c;
 	int i, nbl, nbh;
 	char low[MB_LEN_MAX];
+	mbstate_t mbss;
 
 	nbh = mbtowc (&c, from, MB_CUR_MAX);
 	c = towupper(c);
-	nbl = wctomb_check(low, c);
+	nbl = wcrtomb(NULL, L'\0', &mbss);
+	nbl = wctomb_check(low, c, &mbss);
 
 	/* Check for error on an in-place copy */
 	if ((nbh < nbl) && (to == from))
@@ -532,7 +536,7 @@ char * dictionary_get_data_dir(void)
 }
 
 /**
- * dictopen() - open a dictionary
+ * object_open() -- dictopen() - open a dictionary
  *
  * This function is used to open a dictionary file or a word file,
  * or any associated data file (like a post process knowledge file).
@@ -547,7 +551,9 @@ char * dictionary_get_data_dir(void)
  * sequence of directories is specified in a dictpath string, in
  * which each directory is followed by a ":".
  */
-FILE *dictopen(const char *filename, const char *how)
+void * object_open(const char *filename, 
+                   void * (*opencb)(const char *, void *),
+                   void * user_data)
 {
 	char completename[MAX_PATH_NAME+1];
 	char fulldictpath[MAX_PATH_NAME+1];
@@ -555,7 +561,7 @@ FILE *dictopen(const char *filename, const char *how)
 	static int first_time_ever = 1;
 	char *pos, *oldpos;
 	int filenamelen, len;
-	FILE *fp;
+	void *fp;
 
 	/* Record the first path ever used, so that we can recycle it */
 	if (first_time_ever)
@@ -576,7 +582,7 @@ FILE *dictopen(const char *filename, const char *how)
 	if ((filename[0] == '/') || ((filename[1] == ':') && (filename[2] == '\\'))) 
    {
 		/* fopen returns NULL if the file does not exist. */
-		fp = fopen(filename, how);
+		fp = opencb(filename, user_data);
 		if (fp) return fp;
 	}
 
@@ -624,7 +630,7 @@ FILE *dictopen(const char *filename, const char *how)
 #ifdef _DEBUG
 		prt_error("Info: dictopen() trying %s\n", completename);
 #endif
-		if ((fp = fopen(completename, how)) != NULL) {
+		if ((fp = opencb(completename, user_data)) != NULL) {
 			return fp;
 		}
 		oldpos = pos+1;
@@ -632,6 +638,18 @@ FILE *dictopen(const char *filename, const char *how)
 	return NULL;
 }
 
+static void * dict_file_open(const char * fullname, void * user_data)
+{
+	const char * how = user_data;
+	return fopen(fullname, how);
+}
+
+FILE *dictopen(const char *filename, const char *how)
+{
+	void * ud = (void *) how;
+	return object_open(filename, dict_file_open, ud);
+}
+
 /* ======================================================== */
 /*
  * A static array of precomputed random values. This is used in a
