@@ -25,15 +25,18 @@ static const char * get_a_word(Dictionary dict, FILE * fp)
 	char word[MAX_WORD+4]; /* allow for 4-byte wide chars */
 	const char * s;
 	wint_t c;
+	mbstate_t mbss;
 	int j;
+
 	do {
 		c = fgetwc(fp);
 	} while ((c != WEOF) && iswspace(c));
 	if (c == WEOF) return NULL;
 
+	j = wcrtomb(NULL, L'\0', &mbss);
 	for (j=0; (j <= MAX_WORD-1) && (!iswspace(c)) && (c != WEOF);)
 	{
-		j += wctomb_check(&word[j], c);
+		j += wctomb_check(&word[j], c, &mbss);
 		c = fgetwc(fp);
 	}
 
